package main

import (
	"log"
	"math"
	"math/rand"
	"sync"
)

var EvolveRules = 3
var EvolveCoefs = 2
var EvolveStepsDelta = 5

var StepsYearMultiplier = (yearEnd - yearStart)
var generationsSize = 100
var generations []Generation

func percentile(a float64) float64 {
	var lower float64
	for _, g := range generations {
		// negative fitness
		if g.fitness > a {
			lower++
		}
	}
	return lower / float64(len(generations))
}

func mutationMagnitude(alpha float64) float64 {
	return math.Exp(1.0 / (alpha - 1.01))
}

type Generation struct {
	workingArea  PopulationMap
	expectedArea PopulationMap
	width        int
	height       int
	fitnessFunc  func(g *Generation) float64
	fitness      float64
	rules        Rules
}

func (g *Generation) Fitness() float64 {
	return g.fitnessFunc(g)
}

func (g *Generation) Process() {
	width := g.width
	height := g.height
	next := make([]byte, width*height)
	for y := 0; y < height; y++ {
		for x := 0; x < width; x++ {
			next[y*width+x] = g.workingArea.area[y*width+x]

			if g.workingArea.area[y*width+x] == Border {
				next[y*width+x] = Border
				continue
			}

			count := neighbourCount(g.workingArea.area, width, height, x, y)

			var res float64
			var sum float64
			for i := range Values {
				res += float64(float64(count[i]) * g.rules.koefs[i])
				sum += count[i]
			}

			res /= sum

			for i := range Values {
				if int(g.workingArea.area[y*width+x]) != Values[i] {
					continue
				}

				if res > g.rules.rules[i].evolve {
					if i+1 == len(Values) {
						log.Fatal(res)
					}
					next[y*width+x] = byte(Values[i+1])
				} else if res < g.rules.rules[i].degrade {
					if i == 0 {
						log.Fatal("Less 0", g.rules.rules, g.rules.koefs, res)
					}
					next[y*width+x] = byte(Values[i-1])
				}
			}
		}
	}
	g.workingArea.area = next
}

func (g *Generation) Mutate() {
	cEvolveRules := EvolveRules
	cEvolveCoefs := EvolveCoefs
	cEvolveStepsDelta := EvolveStepsDelta

	for i := -1; i < rand.Intn(cEvolveRules); i++ {
		pos := rand.Intn(len(g.rules.rules))
		// Decrease with randomization
		delta := 2 * (rand.Float64() - 0.5) * mutationMagnitude(percentile(g.fitness))

		what := rand.Intn(2)
		if pos == len(g.rules.koefs)-1 {
			what = 0
		}
		if pos == 0 {
			what = 1
		}

		if what == 0 {
			g.rules.rules[pos].degrade += float64(delta)
		} else {
			g.rules.rules[pos].evolve += float64(delta)
		}

		if g.rules.rules[pos].degrade < -1 && pos != 0 {
			g.rules.rules[pos].degrade = -1
		}
		if g.rules.rules[pos].evolve > 1 && pos != len(g.rules.koefs)-1 {
			g.rules.rules[pos].evolve = 1
		}
		if g.rules.rules[pos].degrade > 1 {
			g.rules.rules[pos].degrade = 1
		}
		if g.rules.rules[pos].evolve < -1 {
			g.rules.rules[pos].evolve = -1
		}
	}

	for i := 0; i < rand.Intn(cEvolveCoefs); i++ {
		pos := rand.Intn(len(g.rules.koefs))
		// Decrease with randomization
		delta := 2 * (rand.Float64() - 0.5) * mutationMagnitude(percentile(g.fitness))

		g.rules.koefs[pos] += float64(delta)
		if g.rules.koefs[pos] > 1 {
			g.rules.koefs[pos] = 1
		}
		if g.rules.koefs[pos] < -1 {
			g.rules.koefs[pos] = -1
		}
	}

	g.rules.steps += rand.Intn(2*cEvolveStepsDelta) - cEvolveStepsDelta + 1
	if g.rules.steps < 1 {
		g.rules.steps = 1
	}
}

func CrossoverRandom(g1 Generation, g2 Generation) (Generation, Generation) {
	g12 := Generation{
		expectedArea: g1.expectedArea,
		width:        g1.width,
		height:       g1.height,
		fitnessFunc:  g1.fitnessFunc,
	}
	g21 := g12

	g12.rules.steps = (g1.rules.steps + g2.rules.steps) / 2
	g21.rules.steps = (g1.rules.steps + g2.rules.steps) / 2

	g12.rules.koefs = make([]float64, len(g1.rules.koefs))
	g21.rules.koefs = make([]float64, len(g1.rules.koefs))

	g12.rules.rules = make([]Rule, len(g1.rules.rules))
	g21.rules.rules = make([]Rule, len(g1.rules.rules))

	for k := 0; k < len(g1.rules.koefs); k++ {
		if rand.Int()%2 == 0 {
			g21.rules.koefs[k] = g1.rules.koefs[k]
			g12.rules.koefs[k] = g2.rules.koefs[k]
		} else {
			g21.rules.koefs[k] = g2.rules.koefs[k]
			g12.rules.koefs[k] = g1.rules.koefs[k]
		}
	}

	for k := 0; k < len(g1.rules.rules); k++ {
		if rand.Int()%2 == 0 {
			g21.rules.rules[k] = g1.rules.rules[k]
			g12.rules.rules[k] = g2.rules.rules[k]
		} else {
			g21.rules.rules[k] = g2.rules.rules[k]
			g12.rules.rules[k] = g1.rules.rules[k]
		}
	}

	return g12, g21
}

func CrossoverAverage(g1 Generation, g2 Generation) (Generation, Generation) {
	g12 := Generation{
		expectedArea: g1.expectedArea,
		width:        g1.width,
		height:       g1.height,
		fitnessFunc:  g1.fitnessFunc,
	}
	g21 := g12

	g12.rules.steps = (g1.rules.steps + g2.rules.steps) / 2
	g21.rules.steps = (g1.rules.steps + g2.rules.steps) / 2

	g12.rules.koefs = make([]float64, len(g1.rules.koefs))
	g21.rules.koefs = make([]float64, len(g1.rules.koefs))

	g12.rules.rules = make([]Rule, len(g1.rules.rules))
	g21.rules.rules = make([]Rule, len(g1.rules.rules))

	for k := 0; k < len(g1.rules.koefs); k++ {
		g21.rules.koefs[k] = (g1.rules.koefs[k] + g2.rules.koefs[k]) / 2
		g12.rules.koefs[k] = (g1.rules.koefs[k] + g2.rules.koefs[k]) / 2
	}

	for k := 0; k < len(g1.rules.rules); k++ {
		g21.rules.rules[k].evolve = (g1.rules.rules[k].evolve + g2.rules.rules[k].evolve) / 2
		g21.rules.rules[k].degrade = (g1.rules.rules[k].degrade + g2.rules.rules[k].degrade) / 2
		g21.rules.rules[k].evolve = (g1.rules.rules[k].evolve + g2.rules.rules[k].evolve) / 2
		g21.rules.rules[k].degrade = (g1.rules.rules[k].degrade + g2.rules.rules[k].degrade) / 2
	}

	return g12, g21
}

func CrossoverSplit(g1 Generation, g2 Generation) (Generation, Generation) {
	g12 := Generation{
		expectedArea: g1.expectedArea,
		width:        g1.width,
		height:       g1.height,
		fitnessFunc:  g1.fitnessFunc,
	}
	g21 := g12

	g12.rules.steps = (g1.rules.steps + g2.rules.steps) / 2
	g21.rules.steps = (g1.rules.steps + g2.rules.steps) / 2

	g12.rules.koefs = make([]float64, len(g1.rules.koefs))
	g21.rules.koefs = make([]float64, len(g1.rules.koefs))

	g12.rules.rules = make([]Rule, len(g1.rules.rules))
	g21.rules.rules = make([]Rule, len(g1.rules.rules))

	for k := 0; k < len(g1.rules.koefs); k++ {
		if k < len(g1.rules.koefs)/2 {
			g12.rules.koefs[k] = g1.rules.koefs[k]
			g21.rules.koefs[k] = g2.rules.koefs[k]
		} else {
			g12.rules.koefs[k] = g2.rules.koefs[k]
			g21.rules.koefs[k] = g1.rules.koefs[k]
		}
	}

	for k := 0; k < len(g1.rules.rules); k++ {
		if k < len(g1.rules.rules)/2 {
			g12.rules.rules[k].evolve = g1.rules.rules[k].evolve
			g21.rules.rules[k].evolve = g2.rules.rules[k].evolve
			g12.rules.rules[k].degrade = g1.rules.rules[k].degrade
			g21.rules.rules[k].degrade = g2.rules.rules[k].degrade
		} else {
			g12.rules.rules[k].evolve = g2.rules.rules[k].evolve
			g21.rules.rules[k].evolve = g1.rules.rules[k].evolve
			g12.rules.rules[k].degrade = g2.rules.rules[k].degrade
			g21.rules.rules[k].degrade = g1.rules.rules[k].degrade
		}
	}

	return g12, g21
}

type Generations struct {
	original PopulationMap
	expected PopulationMap
	gg       []Generation
	min      int
}

func (gg *Generations) Generate() {
	gg.Reset()
	generations = gg.gg

	var wg sync.WaitGroup
	wg.Add(len(gg.gg))
	for i := range gg.gg {
		go func(t int) {
			if mustGenerate {
				gg.gg[t].Mutate()
			}

			for y := 0; y < StepsYearMultiplier*gg.gg[t].rules.steps; y++ {
				gg.gg[t].Process()
			}
			gg.gg[t].fitness = gg.gg[t].Fitness()
			wg.Done()
		}(i)
	}
	wg.Wait()

	m := make(map[int]int)
	free := make(map[int]int)
	for i := 0; i < generationsSize; i += 2 {
		if gg.gg[i].fitness > gg.gg[gg.min].fitness {
			gg.min = i
		}
		if gg.gg[i+1].fitness > gg.gg[gg.min].fitness {
			gg.min = i + 1
		}
		if gg.gg[i].fitness < gg.gg[i+1].fitness {
			gg.gg[i].fitness = 1
			m[i+1] = 1
			free[i] = 1
		} else {
			gg.gg[i+1].fitness = 1
			m[i] = 1
			free[i+1] = 1
		}
	}

	last := -1
	for k := range m {
		if last == -1 {
			last = k
		} else {
			a := -1
			b := -1
			for key := range free {
				if a == -1 {
					a = key
				} else {
					b = key
					break
				}
			}

			delete(free, a)
			delete(free, b)
			gg.gg[a], gg.gg[b] = CrossoverSplit(gg.gg[last], gg.gg[k])

			if gg.min != k && gg.min != last {
				gg.gg[k], gg.gg[last] = CrossoverSplit(gg.gg[last], gg.gg[k])
			}

			last = -1
		}
	}
}

func (gg *Generations) Reset() {
	for i := range gg.gg {
		gg.gg[i].workingArea = gg.original.Copy()
		gg.gg[i].expectedArea = gg.expected
	}
}

func NewGenerations(orig, expected PopulationMap, fitnessFunc func(g *Generation) float64) Generations {
	gg := Generations{
		original: orig,
		expected: expected,
		gg:       make([]Generation, generationsSize),
	}

	for i := range gg.gg {
		gg.gg[i].rules = ReadRules(InputRules)
		gg.gg[i].width = screenWidth
		gg.gg[i].height = screenHeight
		gg.gg[i].fitnessFunc = fitnessFunc
	}

	return gg
}

func neighbourCount(a []byte, width, height, x, y int) []float64 {
	count := make([]float64, len(Values))
	for j := -1; j <= 1; j++ {
		for i := -1; i <= 1; i++ {
			if i == 0 && j == 0 {
				continue
			}
			x2 := x + i
			y2 := y + j
			if x2 < 0 || y2 < 0 || width <= x2 || height <= y2 {
				continue
			}

			for i := range Values {
				if a[y2*width+x2] == byte(Values[i]) {
					count[i] += sqrtOfTwo / 2
					if (x+y)%2 == 0 {
						count[i] += 1 - sqrtOfTwo/2
					}
				}
			}
		}
	}
	return count
}

var sqrtOfTwo = math.Sqrt(2)
