/*
* Version: 1.0.0 - 20.01.2022
* Article: A cellular automaton model for population density dynamics: the case of Ukrainian provinces
*
* Author: Lohachov Mykhailo
*
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
* NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
* LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
* OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
* WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package main

import (
	"fmt"
	_ "image/png"
	"log"
	"math/rand"
	"os"
	"time"

	"github.com/hajimehoshi/ebiten/v2"
)

var (
	IterationsPerYear = 100

	yearStart    = 10
	yearEnd      = 13
	InputRules   = "rules.txt"
	OutputFormat = "out-%d-%d.txt"

	mustGenerate = false
)

// 0 0.5 1 1.5 2 2.5 3 3.5 4 4.5 5
var Values = []int{0, 1, 2, 3, 4, 5}

type World struct {
	Generations
}

func NewWorld() *World {
	area1 := loadOneAllMap(fmt.Sprintf("%s-%d.png", currentCity, yearStart))
	area2 := loadOneAllMap(fmt.Sprintf("%s-%d.png", currentCity, yearEnd))

	w := &World{
		NewGenerations(area1, area2, RRMSEFitness),
	}

	return w
}

var curr = 0

func (w *World) Update() bool {
	if curr < IterationsPerYear {
		w.Generations.Generate()
		curr++
		return true
	}
	w.dumpResults()
	curr = 0

	yearStartID++
	yearEndID++
	if yearEndID == len(years) {
		currentCityID++
		yearStartID = 0
		yearEndID = 1
		if currentCityID == len(cities) {
			os.Exit(0)
		}
	}

	yearStart = years[yearStartID]
	yearEnd = years[yearEndID]
	currentCity = cities[currentCityID]

	area1 := loadOneAllMap(fmt.Sprintf("%s-%d.png", currentCity, yearStart))
	area2 := loadOneAllMap(fmt.Sprintf("%s-%d.png", currentCity, yearEnd))
	w.Generations = NewGenerations(area1, area2, RRMSEFitness)
	return false
}

var (
	screenWidth  = 584
	screenHeight = 489
)

type Game struct {
	world  *World
	pixels []byte
}

var cities []string = []string{
	// "donetsk",
	// "odesa",
	// "cherkasy",
	// "chernivtsy",
	// "chernygiv",
	// "kherson",
	// "kyiv",
	// "ivano-frankivsk",
	// "khmelnitsky",
	// "kirovograd",
	// "mykolaiv",
	// "odessa",
	// "poltava",
	// "rivne",
	// "sumy",
	// "ternopil",
	// "vinnytsa",
	// "volyn",
	// "zakarpattya",
	// "zaporizhzha",
	// "zhitomir",
}

var years []int = []int{10, 13, 16, 19}
var yearStartID = 0
var yearEndID = 1
var currentCityID = 0
var currentCity = cities[currentCityID]

func (g *Game) Update() error {
	if !g.world.Update() {
		g.pixels = nil
		time.Sleep(time.Minute)
	}
	return nil
}

func (g *Game) Draw(screen *ebiten.Image) {
	if g.pixels == nil {
		g.pixels = make([]byte, screenWidth*screenHeight*4)
	}
	g.world.Draw(g.pixels)
	screen.ReplacePixels(g.pixels)
}

func (g *Game) Layout(outsideWidth, outsideHeight int) (int, int) {
	return screenWidth, screenHeight
}

func main() {
	rand.Seed(time.Now().UnixNano())
	g := &Game{
		world: NewWorld(),
	}

	ebiten.SetWindowSize(screenWidth, screenHeight)
	ebiten.SetWindowTitle("CA")
	if err := ebiten.RunGame(g); err != nil {
		log.Fatal(err)
	}
}
